module TicketAbstractorClient
  module ServiceNow
    class Ticket < Base::Ticket
      attr_accessor :sys_id

      def self.fetch_by_id(opts)
        fetch(:by_id, opts).first
      end

      def self.fetch_all_tickets(opts)
        fetch(:all, opts)
      end

      def self.fetch_tickets_by_query(opts)
        fetch(:by_query, opts)
      end

      def initialize(opts = {})
        super(opts)

        @project = opts[:project] || raise(Errors::TicketArgumentError, 'Project is not given')
        @sys_id = @fields['sys_id']
        @status = @fields['state']
      end

      def add_attachment(attachment)
        @attachments << Attachment.new(attachment.merge(ticket_id: @ticket_id, endpoint: @endpoint, project: @project))
        @changes[:new_attachments] += 1

        self
      end

      def add_comment(comment)
        @comments << Comment.new(comment.merge(ticket_id: @ticket_id, endpoint: @endpoint, project: @project))
        @changes[:new_comments] += 1

        self
      end

      def attachments
        return @attachments if @attachments.present?

        return [] if @ticket_id.blank? && @attachments.blank?

        @attachments, attachments_communications_stack = Attachment.fetch(@ticket_id, @endpoint, @project)
        @communications_stack += attachments_communications_stack

        @attachments
      end

      def comments
        return @comments if @comments.present?

        return [] if @ticket_id.blank? && @comments.blank?

        @comments, comments_communications_stack = Comment.fetch(@ticket_id, @endpoint, @project)
        @communications_stack += comments_communications_stack

        @comments
      end

      def to_hash
        { sys_id: @sys_id }.merge(super())
      end

      def sync!
        raise(Errors::TicketArgumentError, 'No changes to apply') unless self.any_changes?

        return create_ticket if @changes[:create]

        update_ticket if @changes[:update]
        update_status if @changes[:new_status]
        sync_comments unless @changes[:new_comments].zero?
        sync_attachments unless @changes[:new_attachments].zero?
        reset_changes!

        self
      end

      def updated_at
        @fields['sys_updated_on']
      end

      private

      # selektor -- all, by_id, by_query
      # Ticket.fetch(:all, opts)
      # Ticket.fetch(:by_id, opts)
      # Ticket.fetch(:by_query, opts)
      def self.fetch(selektor, opts)
        opts = opts.with_indifferent_access
        endpoint = opts.delete(:endpoint)
        method_map = { all: :get_all_tickets, by_id: :get_ticket_by_id, by_query: :get_tickets_by_query }
        client = Client.new(endpoint)
        client_response = client.public_send(method_map.fetch(selektor.to_sym), opts)

        Array.wrap(client_response).map do |ticket_hash|
          ticket_id = opts[:ticket_id] || ticket_hash['number']
          ticket = new(ticket_id: ticket_id, fields: ticket_hash, endpoint: endpoint, project: opts[:project])
          filtered_ticket = filter_ticket(ticket)
          filtered_ticket.reset_changes!
          filtered_ticket.communications_stack = client.communications_stack
          filtered_ticket
        end
      end

      def self.filter_ticket(raw_ticket)
        tickets_filter_class = TicketAbstractorClient.configuration.snow_tickets_filter_class

        return raw_ticket if tickets_filter_class.blank?

        tickets_filter_class.new(raw_ticket).filter_ticket
      end

      def create_ticket
        client = Client.new(@endpoint)
        response = client.create_ticket(self)
        @communications_stack += client.communications_stack

        ticket = Ticket.fetch_by_id(ticket_id: response['number'], project: @project, endpoint: @endpoint)

        @sys_id = ticket.sys_id
        @fields = ticket.fields
        @status = ticket.status
        @ticket_id = ticket.ticket_id
        @communications_stack += ticket.communications_stack
        reset_changes!

        self
      end

      def update_status
        @fields['state'] = @status
        update_ticket
      end

      def update_ticket
        client = Client.new(@endpoint)
        response = client.update_ticket(self)
        @communications_stack += client.communications_stack
        @fields['sys_updated_on'] = response['sys_updated_on']
      end

      def sync_attachments
        @communications_stack +=
          @attachments.last(@changes[:new_attachments]).flat_map do |attachment|
            attachment.sync!
            attachment.communications_stack
          end
      end

      def sync_comments
        @communications_stack +=
          @comments.last(@changes[:new_comments]).flat_map do |comment|
            comment.sync!
            comment.communications_stack
          end
      end
    end
  end
end
